module Hoare (
  wp,
  verify
  ) where

import State
import Imp
import Prover

-- generate weakest precondition
-- wp f c generates a pair of the weakest precondition for c with respect to 
-- postcondition f and a list of auxiliary predicates to prove
wp :: BExp -> Comm -> (BExp, [BExp])
wp f Skip            = (f,[])
wp f (x := e)        = (error "Write Me!", [])  -- implement the backwards assignment 
wp f (c1 :> c2)      =                          -- rule here using a substitution 
  let (f',fs')   = wp f c2 in 
  let (f'',fs'') = wp f' c1 in
  (f'', fs' ++ fs'')
wp f (If b c1 c2)    = 
  let (f1, fs1) = wp f c1 
      (f2, fs2) = wp f c2
  in (((b :==> f1) :&: (Not b :==> f2)), fs1 ++ fs2)
wp f (While b inv c) = 
  let (f1, fs1) = wp inv c in
  (inv, (inv :&: b :==> f1) : (inv :&: Not b :==> f) : fs1)

-- verify p prog q is True when the Hoare Triple { p } c { q } is valid
--   where prog = Prog ds c 
--   verify must check that the precondition p implies wp f c
--   verify must also check that p, c, and q are well-formed with respect to ds
--     if they are well-formed, proceed.
--     if they are not well-formed, use illFormedTriple to signal an error

illFormedTriple = error "Ill-formed Hoare Triple"

verify :: BExp -> Program -> BExp -> Bool
verify p (Prog ds c) q =
  if goodB ds p && good (Prog ds c) && goodB ds q then 
    let (f,fs) = wp q c in
    all (\f -> prove f ds) ((p :==> f) : fs)
  else
    illFormedTriple